%% Connexion with WordNet
:- wn_connect.

%% For this problem, where we mainly unify words (constans) Sessa's weak unification
%% algorithm (a1) is enough and less expensive  
:- weak_unification(a1). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% all_concepts(+Type, -List_of_concepts, -Num_of_concepts)

all_concepts(Type, List_of_concepts, Num_of_concepts) :-
    %% synset type:
    %% n (noun); v (verb); a (adjective); s (satellite adjective) and r (adverb).
    member(Type, [n,v,a,s,r]),
	findall((Word:Type:Sense), wn_s(_, _, Word, Type, Sense, _), List_of_concepts),
    length(List_of_concepts, Num_of_concepts).


%%% all_concepts(+Type, -List_of_concepts)
all_concepts(Type, List_of_concepts) :-
%% synset type:
%% n (noun); v (verb); a (adjective); s (satellite adjective) and r (adverb).
member(Type, [n,v,a,s,r]),
findall((Word:Type:Sense), wn_s(_, _, Word, Type, Sense, _), List_of_concepts).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% take_one_pair_from_list(+List, -[Element1, Element2])
%%% Takes one pair of element from the list List.
take_one_pair_from_list(List, [Element1, Element2]) :-
    random_member(Element1, List),
    random_member(Element2, List).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% gen_list_of_pairs(+Num_of_pairs, -List_of_pairs)
%% Generates a ramdom list of pairs from two lists (the list of all nouns
%% and the list of all verbs in WordNet).
gen_list_of_pairs(Num_of_pairs, List_of_pairs) :-
    ((Num_of_pairs < 0; (Num_of_pairs mod 2) =\= 0) ->
        write('Enter an even integer greater than 0')
        ;
        Num is Num_of_pairs div 2,
        all_concepts(n, List_of_nouns),
        all_concepts(v, List_of_verbs),
        gen_list_of_pairs(List_of_nouns, List_of_verbs, Num, [], List_of_pairs)
    ).

gen_list_of_pairs(_, _, 0, List_of_pairs, List_of_pairs):- !.
gen_list_of_pairs(List_of_nouns, List_of_verbs, Num_of_pairs, List_of_pairs_Acc, List_of_pairs):-
    take_one_pair_from_list(List_of_nouns, [Noun1, Noun2]),
    take_one_pair_from_list(List_of_verbs, [Verb1, Verb2]),
    Num is Num_of_pairs - 1,
    gen_list_of_pairs(List_of_nouns, List_of_verbs, Num, [[Noun1, Noun2],[Verb1, Verb2]|List_of_pairs_Acc], List_of_pairs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% store_list_of_pairs(+Num_of_pairs, +Filename)
store_list_of_pairs(Num_of_pairs, Filename):-
    (exists_file(Filename) ->
        write('The file '), write(Filename), write(' does exists.'), nl
        ;
        %% generates a list of Num_of_pairs pairs
        gen_list_of_pairs(Num_of_pairs, List_of_pairs),
        %% saves the list
        open(Filename, write, OutputStream),
        write(OutputStream,'['),
        save_list_of_pairs_in(List_of_pairs, OutputStream),
        write(OutputStream,']'),
        write(OutputStream,'.'),
        nl(OutputStream),
        close(OutputStream)
    ).

%% save_list_of_pairs_in(List_of_pairs, OutputStream)
save_list_of_pairs_in([], _).
save_list_of_pairs_in([[C1, C2]], OutputStream) :-
        !,
        write(OutputStream,'['),
        format(OutputStream,'~q', [C1]),
        write(OutputStream,','),
        format(OutputStream,'~q', [C2]),
        write(OutputStream,']').

save_list_of_pairs_in([[C1, C2]|List_of_pairs], OutputStream) :-
        write(OutputStream,'['),
        format(OutputStream,'~q', [C1]),
        write(OutputStream,','),
        format(OutputStream,'~q', [C2]),
        write(OutputStream,']'),
        write(OutputStream,', '),
        save_list_of_pairs_in(List_of_pairs, OutputStream).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% get_list_of_pairs(+Filename, -List_of_pairs)
%% reads a list of pairs stored in Filename
get_list_of_pairs(Filename, List_of_pairs) :-
(exists_file(Filename) ->
    see(Filename),
    read(List_of_pairs),
    seen
    ;
    write('The file '), write(Filename), write(' does not exists.'), nl
).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% benchmark(+List_of_pairs, +Measure)
benchmark(List_of_pairs, Measure):-
    length(List_of_pairs, Num_of_pairs),
    stat_start(I),
    ((Measure=path)-> bench_path(List_of_pairs);
    (Measure=wup)  -> bench_wup(List_of_pairs);
    (Measure=lch)  -> bench_lch(List_of_pairs);
    (Measure=res)  -> bench_res(List_of_pairs);
    (Measure=jcn)  -> bench_jcn(List_of_pairs);
    (Measure=lin)  -> bench_lin(List_of_pairs)),
    stat_end(+I, Time, Inferences),
    %% Latency = Time in msec/Num_of_pairs = (msec/pair)
    Latency is (Time*1000)/Num_of_pairs,
    %% Throughput = Num_of_pairs/Time in sec = (pairs/sec)
    Throughput is Num_of_pairs/Time,
    write('Measure '), write(Measure), write(' :'),
    write('  Time='), write(Time), write(' (sec); '),
    write('  Latency='), write(Latency), write(' (msec/pair); '),
    write('  Throughput='), write(Throughput), write(' (pairs/sec); '),
    write('  Inferences='), write(Inferences), nl.


%%% bench_???(+List_of_pairs)
bench_path([]).
bench_path([[C1,C2]|Pairs]) :- wn_path(C1,C2,_D), bench_path(Pairs).

bench_wup([]).
bench_wup([[C1,C2]|Pairs]) :- wn_wup(C1,C2,_D), bench_wup(Pairs).

bench_lch([]).
bench_lch([[C1,C2]|Pairs]) :- wn_lch(C1,C2,_), bench_lch(Pairs).

bench_res([]).
bench_res([[C1,C2]|Pairs]) :- wn_res(C1,C2,_), bench_res(Pairs).

bench_jcn([]).
bench_jcn([[C1,C2]|Pairs]) :- wn_jcn(C1,C2,_), bench_jcn(Pairs).

bench_lin([]).
bench_lin([[C1,C2]|Pairs]) :- wn_lin(C1,C2,_), bench_lin(Pairs).



experiment(Filename, Measure) :- get_list_of_pairs(Filename, List_of_pairs), benchmark(List_of_pairs, Measure).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% benchmark2(+List_of_pairs, +Measure)
benchmark2(List_of_pairs, Measure):-
    ((Measure=path)-> bench_path(List_of_pairs);
     (Measure=wup)  -> bench_wup(List_of_pairs);
     (Measure=lch)  -> bench_lch(List_of_pairs);
     (Measure=res)  -> bench_res(List_of_pairs);
     (Measure=jcn)  -> bench_jcn(List_of_pairs);
     (Measure=lin)  -> bench_lin(List_of_pairs)
	).

experiment2(Filename, Measure) :- get_list_of_pairs(Filename, List_of_pairs), time(benchmark2(List_of_pairs, Measure)).


%%%%%%%%%%%%%%
%%       STATISTICS
%%%%%%%%%%%%%%
% This is how time and space are measured.

% stat_start(-I), Initialization procedure. I is the total number of inferences
%              (passes via the call and redo ports) since the system was started.
% stat_start(output)
stat_start(I) :-
    garbage_collect,
    statistics(runtime,_), statistics(inferences,I).

% stat_end(+I, Time, Inferences),
% "I" must be the total number of inferences computed by stat_start
% stat_end(input)
stat_end(I, Time, Inferences) :-
    %% Runtime in seconds
    statistics(runtime,[_,MSec]), Time is MSec/1000,
    statistics(inferences,I2), Inferences is I2 - I.


